/**
 * $Author: iyoon $
 * $Date: 2009/10/01 21:56:19 $
 * $Revision: 1.2 $
 * 
 * Written by Il-Chul Yoon <iyoon@cs.umd.edu>
 */
 
#include <iostream>
#include <assert.h>
#include <string.h>
#include "IC_EndPointSet.h"
#include "IC_Connection.h"

const char * const IC_EndPointSet::IC_epset_errors[] = {
    "no error",
    "generic error", 
    "invalid number of array dimensions",
    "can't allocate InterComm array regions",
    "can't obtain distributed array descriptor",
    "can't compute communication schedule",
    "communication (send/recv) failure",
    "invalid endpoint name",
    "this program's initialization failed",
    "can't connect to remote program",
    "PVM error. Is PVM running?",
    "MPI error",
    "internal pointer table error",
    "fail to initialize endpoint set from XJD",
    "fail to register array with array name",
    "fail to commit the arrays",
};

IC_EndPointSet::IC_EndPointSet(const char* xjdfile, const char* compname, int& stat) {

    std::cout << "arguments :" << XJDVERSION << " " << xjdfile << " " << compname << " " << stat << "\n";

    ezxml_t xmlRoot = read_xjd(xjdfile);

    this->myprogname = compname;
    this->xjdversion = strdup(get_version(xmlRoot));
    if(this->xjdversion == NULL){
      fprintf(stderr, "XJD version is missing...\n");
      stat = -1;
    }
    else{
      fprintf(stderr, "version %s\n", this->xjdversion);
      
      assert(XJD_Validate(xmlRoot) == 0);
      
      if(!strcmp(this->xjdversion, "1.5")){
        epset_init_1_5(xmlRoot, stat);
      }
      else if(!strcmp(this->xjdversion, "2.0")){
        epset_init_2_0(xmlRoot, stat);
      }
    }

    assert(stat == 0);
}

IC_EndPointSet::~IC_EndPointSet(){
  map<string, IC_Connection*>::const_iterator citer; 
  for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
    string connname = citer->first;    
    IC_Connection* conn = (IC_Connection*)citer->second;
    delete conn;
  }

  map<string, IC_EndPoint*>::const_iterator eiter; 
  for (eiter = this->epmap.begin(); eiter != this->epmap.end(); ++eiter) {
    IC_EndPoint* ep = (IC_EndPoint*)eiter->second;
    delete ep;
  }

  /* clear maps*/
  epmap.clear();
  connmap.clear();

#ifdef PPLUSPLUS
    /* clear arrays */
    ppp_int_arrmap.clear();
    ppp_float_arrmap.clear();
    ppp_double_arrmap.clear();
#endif

#ifdef FORTRAN90
    /* free the encapsulated arraydesc for each Fortran90 array */
    map<string, const F90_charArray*>::const_iterator f90_char_arrdesc_iter; 
    map<string, const F90_shortArray*>::const_iterator f90_short_arrdesc_iter; 
    map<string, const F90_intArray*>::const_iterator f90_int_arrdesc_iter; 
    map<string, const F90_floatArray*>::const_iterator f90_float_arrdesc_iter; 
    map<string, const F90_doubleArray*>::const_iterator f90_double_arrdesc_iter; 
    /* the arrays for the broadcast is not registered and the array itself 
       is used to create corresponding Array meta info (arraydesc) when 
       bcast call is invoked. After bcast, the meta info will be destroyed.

       So, the f90_type_arrmap has the port names for the broadcast, but,
       the value would be null (initialized during the XJD initialization)

       Here, we delete only the meta info structure such as F90_intArray, 
       which was mapped to the port name before.
    */

    for (f90_char_arrdesc_iter = this->f90_char_arrmap.begin(); f90_char_arrdesc_iter != this->f90_char_arrmap.end(); ++f90_char_arrdesc_iter) {
        F90_charArray* a = (F90_charArray*)f90_char_arrdesc_iter->second;
        if(a != NULL)
          delete a;
    }

    for (f90_short_arrdesc_iter = this->f90_short_arrmap.begin(); f90_short_arrdesc_iter != this->f90_short_arrmap.end(); ++f90_short_arrdesc_iter) {
        F90_shortArray* a = (F90_shortArray*)f90_short_arrdesc_iter->second;
        if(a != NULL) 
          delete a;
    }

    for (f90_int_arrdesc_iter = this->f90_int_arrmap.begin(); f90_int_arrdesc_iter != this->f90_int_arrmap.end(); ++f90_int_arrdesc_iter) {
        F90_intArray* a = (F90_intArray*)f90_int_arrdesc_iter->second;
        if(a != NULL)
          delete a;
    }

    for (f90_float_arrdesc_iter = this->f90_float_arrmap.begin(); f90_float_arrdesc_iter != this->f90_float_arrmap.end(); ++f90_float_arrdesc_iter) {
        F90_floatArray* a = (F90_floatArray*)f90_float_arrdesc_iter->second;
        if(a != NULL)
          delete a;
    }

    for (f90_double_arrdesc_iter = this->f90_double_arrmap.begin(); f90_double_arrdesc_iter != this->f90_double_arrmap.end(); ++f90_double_arrdesc_iter) {
        F90_doubleArray* a = (F90_doubleArray*)f90_double_arrdesc_iter->second;
        if(a != NULL)
          delete a;
    }

    /* clear arrays	*/
    f90_char_arrmap.clear();
    f90_short_arrmap.clear();
    f90_int_arrmap.clear();
    f90_float_arrmap.clear();
    f90_double_arrmap.clear();
#endif
 
}

int IC_EndPointSet::XJD_Validate(ezxml_t xmlRoot){
  /* return -1 if error in validation, else return 0*/
  if(!strcmp(this->xjdversion, "1.5")){
    if(XJD_Validate_1_5(xmlRoot) < 0){
      return -1;
    }
  }
  else if(!strcmp(this->xjdversion, "2.0")){
    return XJD_Validate_2_0(xmlRoot);
  }

  return 0;
}

int IC_EndPointSet::XJD_Validate_1_5(ezxml_t xmlRoot){
  if(check_multi_import(xmlRoot))
    return -1;

   /** 
    * TODO: should implement later...
    *   inspect for the component and connection information 
    *   based on the version information
    */

   return 0; 
  
}

int IC_EndPointSet::check_multi_import(ezxml_t xmlRoot){
  ezxml_t curConn, tmpConn, exp, imp;
  char* impname;
  char* impport;
  char* tmp_impname;
  char* tmp_impport;
  int n = 0;
    
  curConn = get_first_conn(xmlRoot);
  while(curConn){
    impname = ezxml_child(curConn, IMPORTER)->txt;
    impport = ezxml_child(curConn, IMPPORT)->txt;
      
    tmpConn = get_first_conn(xmlRoot);
    while(tmpConn){
      tmp_impname = ezxml_child(tmpConn, IMPORTER)->txt;
      tmp_impport = ezxml_child(tmpConn, IMPPORT)->txt;
        
      if(!strcmp(impname, tmp_impname) && !strcmp(impport, tmp_impport)){
        n++;
      }
        
      tmpConn = tmpConn->next;
    }
      
    if(n > 1){
      fprintf(stderr, "\tImport from multiple program is detected.\n");
      return 1;
    }
    else{
      n = 0;
      curConn = curConn->next;
    }
  }

  return 0;
}


int IC_EndPointSet::XJD_Validate_2_0(ezxml_t xmlRoot){
  return 1;
}
    
int IC_EndPointSet::epset_init_1_5(ezxml_t xmlRoot, int& stat){
    stat = register_program(xmlRoot);
    std::cout << "Register Status of " << this->myprogname << " : " << stat << "\n";
    assert(stat == IC_OK);

    stat = inspect_connections(xmlRoot);
    assert(stat == IC_OK);

    return 0;
}

/**
 * Precondition: following class variables must be defined before.
 *   const char* myprogname;
 *   unsigned mynproc; 
 *   unsigned myao;
 */
int IC_EndPointSet::inspect_connections(ezxml_t xmlRoot){
  ezxml_t curConn, exp, imp, expport, impport, commtype, datatype, connid;
  char* pname;
  char* pname2;
  char* other_nproc;
  char* other_arrord;
  char* exportport;
  char* importport; 
  int tag = -1;
    
  unsigned onproc = 0;
  unsigned oao = 0;
  int stat = IC_OK;
    
    curConn = get_first_conn(xmlRoot);
    
    while(curConn){
        connid = ezxml_child(curConn, CONNID);
		
        exp = ezxml_child(curConn, EXPORTER);
        pname = get_field_byID(exp->txt, xmlRoot, COMPNAME);
        
        if( !strcmp(this->myprogname, pname) ){
#ifdef IC_DEBUG           
            fprintf(stderr, "ConnID: %s\n", connid->txt);
            fprintf(stderr, "\tappear as exporter: %s\n", pname);
#endif

            imp = ezxml_child(curConn, IMPORTER); /* get cid of partner program*/
            pname2 = get_field_byID(imp->txt, xmlRoot, COMPNAME);            

            other_nproc = get_field_byID(imp->txt, xmlRoot, N_NODE);
            if(other_nproc){
                onproc = (unsigned) atoi(other_nproc);
            }

            other_arrord = get_field_byID(imp->txt, xmlRoot, ARRAYORDER);
            if(other_arrord){
                if (!strcmp(other_arrord, "IC_COLUMN_MAJOR")){
                    oao = IC_EndPoint::IC_COLUMN_MAJOR;
                }
                else if(!strcmp(other_arrord, "IC_ROW_MAJOR")){
                    oao = IC_EndPoint::IC_ROW_MAJOR;
                }
                else{ /* error*/
                    fprintf(stderr, "Wrong array order in XJD.\n");
                    return IC_EPSET_INIT_FAILURE;
                }
            }

            expport = ezxml_child(curConn, EXPPORT);
            commtype = ezxml_child(curConn, COMMTYPE);
            datatype = ezxml_child(curConn, DATATYPE);
            tag = atoi(ezxml_child(curConn, TAG)->txt);
            
            assert( (pname) && (onproc > 0) && 
                    ((oao == IC_EndPoint::IC_COLUMN_MAJOR) || (oao == IC_EndPoint::IC_ROW_MAJOR)) &&
                    (expport != NULL) && (commtype != NULL) && (datatype != NULL)
                  );
                  
            addep(this->myprogname, this->mynproc, this->myao, 
                  pname2, onproc, oao, 
                  expport->txt, commtype->txt, datatype->txt, tag, 0, connid->txt);
                
        }
        
        imp = ezxml_child(curConn, IMPORTER);
        pname = get_field_byID(imp->txt, xmlRoot, COMPNAME);
        
        if( !strcmp(this->myprogname, pname) ){
#ifdef IC_DEBUG           
            fprintf(stderr, "ConnID: %s\n", connid->txt);
            fprintf(stderr, "\tappear as importer: %s\n", pname);
#endif			
            exp = ezxml_child(curConn, EXPORTER);
            pname2 = get_field_byID(exp->txt, xmlRoot, COMPNAME);            

            other_nproc = get_field_byID(exp->txt, xmlRoot, N_NODE);
            if(other_nproc){
                onproc = (unsigned) atoi(other_nproc);
            }

            other_arrord = get_field_byID(exp->txt, xmlRoot, ARRAYORDER);
            if(other_arrord){
                if (!strcmp(other_arrord, "IC_COLUMN_MAJOR")){
                    oao = IC_EndPoint::IC_COLUMN_MAJOR;
                }
                else if(!strcmp(other_arrord, "IC_ROW_MAJOR")){
                    oao = IC_EndPoint::IC_ROW_MAJOR;
                }
                else{ /* error*/
                    fprintf(stderr, "Wrong array order in XJD.\n");
                    return IC_EPSET_INIT_FAILURE;
                }
            }

            impport = ezxml_child(curConn, IMPPORT);
            commtype = ezxml_child(curConn, COMMTYPE);
            datatype = ezxml_child(curConn, DATATYPE);
            tag = atoi(ezxml_child(curConn, TAG)->txt);
            
            assert( (pname) && (onproc > 0) && 
                    ((oao == IC_EndPoint::IC_COLUMN_MAJOR) || (oao == IC_EndPoint::IC_ROW_MAJOR)) &&
                    (impport != NULL) && (commtype != NULL) && (datatype != NULL)
                  );
                  
            addep(this->myprogname, this->mynproc, this->myao,
                  pname2, onproc, oao, 
                  impport->txt, commtype->txt, datatype->txt, tag, 1, connid->txt);
        }
        
        exp = imp = NULL;
        curConn = curConn->next;
    }
    
    return stat;
}


void IC_EndPointSet::addep(const char* myname, unsigned mynproc, unsigned myao,
                           char* othername, unsigned onproc, unsigned oao,
                           char* portname, char* commtype, char* datatype, int tag,
						   int ex_im, char* connname){
    int stat;                            
    string p1 = myname;
    string p2 = othername;
    string epname = p1 + ":" + p2;


#ifdef IC_DEBUG       
    cout << endl;
    cout << "MyInfo: " << myname << " " << mynproc << " " << myao << endl;
    cout << "OtherInfo: " << othername << " " << onproc << " " << oao << endl;
    cout << endl;
#endif
    
    IC_EndPoint* ep = findep(epname);
    if(ep == NULL){
#ifdef IC_DEBUG       
        cout << "\tRegistering new endpoint " << epname << " and port " << portname << "." << endl;
#endif

        ep = new IC_EndPoint(epname.c_str(), mynproc, onproc, myao, oao, stat);
        assert(stat == IC_EndPoint::IC_OK);
        
        this->epmap[epname] = ep;

        registerConn(ep, connname, portname, commtype, datatype, tag, ex_im, stat);
        assert(stat == IC_OK);
    }
    else{
#ifdef IC_DEBUG       
        cout << "\tEndpoint exists! Just register port " << portname << "." << endl;
#endif
        registerConn(ep, connname, portname, commtype, datatype, tag, ex_im, stat);
        assert(stat == IC_OK);
    }
}

IC_EndPoint* IC_EndPointSet::findep(string epname){
    map<string, IC_EndPoint*>::const_iterator iter;

    iter = this->epmap.find(epname);
    if(iter == this->epmap.end()){
#ifdef IC_DEBUG
      cout << "no such endpoint name in endpoint map" << endl;
#endif
        return NULL;
    }
    else{
        return (IC_EndPoint*)iter->second;
    }
}


void IC_EndPointSet::registerConn(IC_EndPoint* ep, char* connname, char* ptname, 
                                  char* ctype, char* dtype, int tag, int ex_im, 
                                  int& stat){
	IC_Connection* conn = new IC_Connection(connname, ptname);
	conn->set_type(ex_im);

  string conntype = ctype;
  conn->set_ctype(conntype);
  
  conn->set_ep(ep);
  conn->set_tag(tag);
  
  string datatype = dtype;
  conn->set_dtype(datatype);

  this->connmap[connname] = conn;

#ifdef IC_DEBUG
	cout << "registering connection " << connname << " as type " << ex_im << endl;
#endif

  //
  // just assign space for the array that will be linked with later.
  //
  string portname = ptname;

#ifdef PPLUSPLUS
  if(!strcmp(dtype, "int")){
    this->ppp_int_arrmap[portname] = NULL;

#ifdef IC_DEBUG
    cout << "\tintArray Map Size is : " << this->ppp_int_arrmap.size() << endl;
#endif
  }
  else if(!strcmp(dtype, "float")){
    this->ppp_float_arrmap[portname] = NULL;
#ifdef IC_DEBUG
    cout << "\tfloatArray Map Size is : " << this->ppp_float_arrmap.size() << endl;
#endif
  }
  else if(!strcmp(dtype, "double")){
    this->ppp_double_arrmap[portname] = NULL; 
#ifdef IC_DEBUG
	cout << "\tdoubleArray Map Size is : " << this->ppp_double_arrmap.size() << endl;
#endif
  }
  else{
    cout << "\tInvalid datatype in XJD" << endl;
    stat = IC_EPSET_INIT_FAILURE;
    return;
  }
#endif

#ifdef FORTRAN90
  if(!strcmp(dtype, "int")){
    this->f90_int_arrmap[portname] = NULL;
#ifdef IC_DEBUG
    cout << "\tintArray Map Size is : " << this->f90_int_arrmap.size() << endl;
#endif
  }
  else if(!strcmp(dtype, "float")){
    this->f90_float_arrmap[portname] = NULL;
#ifdef IC_DEBUG
    cout << "\tfloatArray Map Size is : " << this->f90_float_arrmap.size() << endl;
#endif
  }
  else if(!strcmp(dtype, "double")){
    this->f90_double_arrmap[portname] = NULL;
#ifdef IC_DEBUG
    cout << "\tfloatArray Map Size is : " << this->f90_double_arrmap.size() << endl;
#endif
  }
  else{
    cout << "\tInvalid datatype in XJD" << endl;
    stat = IC_EPSET_INIT_FAILURE;
    return;
  }
#endif
  
  stat = IC_OK;
}


int IC_EndPointSet::epset_init_2_0(ezxml_t xmlRoot, int& stat){
    return IC_OK;
}



ezxml_t IC_EndPointSet::read_xjd(const char* xjdname){
    FILE* fp; 
    
    if ((fp = fopen(xjdname, "r")) == NULL){
        fprintf(stderr, "Cannot open %s\n", xjdname);
        return NULL;
    }
    return ezxml_parse_fp(fp);
}


char* IC_EndPointSet::get_version(ezxml_t xmlRoot){
  ezxml_t v = ezxml_child(xmlRoot, XJDVERSION);
  if(v != NULL){
    return v->txt;
  }
  else{
    return NULL;
  }
}

/**
 * decide following class variables for defined components in XJD
 *   char* cid;
 *   unsigned mynproc; 
 *   unsigned myao;
 */ 
int IC_EndPointSet::register_program(ezxml_t xmlRoot){
  ezxml_t curComp;
    ezxml_t tmp, tmpstr, cid;
   
    curComp = get_first_comp(xmlRoot);
    
    while(curComp){
      cid = ezxml_child(curComp, COMPID); // set cid
      tmp = ezxml_child(curComp, COMPNAME);
      
      if(cid == NULL){
        fprintf(stderr, "component id is required (Review XJD file)\n");
        return IC_EPSET_INIT_FAILURE;
      }
      if(tmp == NULL){
        fprintf(stderr, "component name is required (Review XJD file)\n");
        return IC_EPSET_INIT_FAILURE;
      }
      
      // same component name must appear once in XJD
      if( !strcmp(this->myprogname, tmp->txt) ){ 
        this->cid = (char*) strdup(cid->txt);
        
        tmpstr = ezxml_child(curComp, N_NODE); 
        if(tmpstr){
          this->mynproc = (unsigned) atoi(tmpstr->txt); // set mynproc	
        }
        
        tmpstr = ezxml_child(curComp, ARRAYORDER); 
        if(tmpstr){
          if (!strcmp(tmpstr->txt, "IC_COLUMN_MAJOR")){
            this->myao = IC_EndPoint::IC_COLUMN_MAJOR; // set myao	
          }
          else if(!strcmp(tmpstr->txt, "IC_ROW_MAJOR")){
            this->myao = IC_EndPoint::IC_ROW_MAJOR; // set myao
            fprintf(stderr, "Warning: P++/F90 array will be handled in column major internally. Recommend to use IC_COLUMN_MAJOR for arrayorder of component %s\n", this->myprogname);
            assert(this->myao == IC_EndPoint::IC_COLUMN_MAJOR);
          }
          else{ /* error*/
            fprintf(stderr, "Wrong array order in XJD.\n");
            return IC_EPSET_INIT_FAILURE;
          }
        }
        
        return IC_OK;
      }
      
      curComp = curComp->next;
    }
    
    return IC_EPSET_INIT_FAILURE;
}


ezxml_t IC_EndPointSet::get_first_comp(ezxml_t xmlRoot){
    ezxml_t plist = ezxml_child(xmlRoot, COMPLIST);
    return ezxml_child(plist, COMP);
}


ezxml_t IC_EndPointSet::get_first_conn(ezxml_t xmlRoot){
    ezxml_t clist = ezxml_child(xmlRoot, CONNLIST);
    return ezxml_child(clist, CONN);
}


char* IC_EndPointSet::get_field_byID(char* id, ezxml_t xmlRoot, char* fieldname){
    ezxml_t curComp;
    ezxml_t cid, tmp;
   
    curComp = get_first_comp(xmlRoot);
    
    while(curComp){
        cid = ezxml_child(curComp, COMPID);
        tmp = ezxml_child(curComp, fieldname);

        if( !strcmp(cid->txt, id) ){
            return tmp->txt;
        }

        curComp = curComp->next;
    }

    return NULL;
}

void IC_EndPointSet::commitArrays(int& stat){
    map<string, IC_Connection*>::const_iterator citer;
    for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
      string connname = citer->first;    
      IC_Connection* conn = (IC_Connection*)citer->second;
      
      if(!strcmp(conn->get_ctype().c_str(), "1xN")){
#ifdef IC_DEBUG
        cout << "1xN connection: " << connname << endl;
#endif
        continue;
      }

      IC_EndPoint* cur_ep = (IC_EndPoint*)conn->get_ep();
      IC_Program* myprog = cur_ep->get_myprog();
      IC_Program* otherprog = cur_ep->get_otherprog();
      IC_Desc*    desc = conn->get_desc();
      IC_Region** region_set = conn->get_rgn();

      // ignore incomplete regions
      if(myprog == NULL || otherprog == NULL || desc == NULL || region_set == NULL){
        continue;
      }

      cout << "\tCommitting connection : " << conn->get_connname() << " : "
           << conn->get_portname() << " : "
           << myprog->name << " " << otherprog->name << endl;

      /* compute schedule for the regionset corresponding to the given portname*/
      /* region set size has been fixed to 1*/
      IC_Sched* sched = IC_Compute_schedule(myprog, otherprog, desc, region_set, 1);
      if (!sched) {
#ifdef IC_DEBUG
        cout << "port : " << conn->get_portname() << " : "
             << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE] 
             << endl;
#endif
        stat = IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE;
        return;
      }
      else{
        conn->set_sched(sched);
      }
    }

    stat = IC_OK;
}

void IC_EndPointSet::exportArray(char* ptname, int& stat){
#ifdef IC_DEBUG
  cout << "Export Port" << endl;
#endif

  int n_exp = 0;
  map<string, IC_Connection*>::const_iterator citer;

  for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
    string connname = citer->first;    
    IC_Connection* conn = (IC_Connection*)citer->second;
      
    if(!strcmp(conn->get_ctype().c_str(), "1xN")){ 
#ifdef IC_DEBUG
      cout << "\tskip 1xN connection: " << connname << endl;
#endif
      continue;
    }

    if(conn->get_portname().compare(ptname) != 0){ // different portname
#ifdef IC_DEBUG
      cout << "\tskip different port name..." << endl;
#endif
      continue;
    }

    if(conn->get_type() != 0){ // connection is not for export
#ifdef IC_DEBUG
      cout << "\tskip port for import data" << endl;
#endif
      continue;
    }
        
    n_exp++;

    IC_EndPoint* ep = (IC_EndPoint*)conn->get_ep();
    IC_Program* myprog = ep->get_myprog();
    IC_Program* otherprog = ep->get_otherprog();
    IC_Sched* sched = conn->get_sched();
    int tag = conn->get_tag();
     
    cout << "\tWaiting to export : " << ptname 
         << " to " << otherprog->name << endl;

#ifdef PPLUSPLUS
    map<string, const intArray*>::const_iterator intarr_iter; 
    intarr_iter = this->ppp_int_arrmap.find(ptname);
    if(intarr_iter != this->ppp_int_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in intArrays" << endl;
#endif

      const intArray* intarr = (const intArray*)intarr_iter->second;
        
      int ret = IC_Send_int(otherprog, sched, 
                            intarr->getLocalArray().getDataPointer(), tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }

    map<string, const floatArray*>::const_iterator floatarr_iter; 
    floatarr_iter = this->ppp_float_arrmap.find(ptname);
    if(floatarr_iter != this->ppp_float_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in floatArrays" << endl;
#endif
        
      const floatArray* floatarr = (const floatArray*)floatarr_iter->second;
        
      int ret = IC_Send_float(otherprog, sched, 
                              floatarr->getLocalArray().getDataPointer(), tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }

    map<string, const doubleArray*>::const_iterator doublearr_iter; 
    doublearr_iter = this->ppp_double_arrmap.find(ptname);
    if(doublearr_iter != this->ppp_double_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in doubleArrays" << endl;
#endif
        
      const doubleArray* doublearr = (const doubleArray*)doublearr_iter->second;
        
      int ret = IC_Send_double(otherprog, sched, 
                               doublearr->getLocalArray().getDataPointer(), tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }
#endif

#ifdef FORTRAN90
    map<string, const F90_charArray*>::const_iterator f90chararr_iter; 
    f90chararr_iter = this->f90_char_arrmap.find(ptname);
    if(f90chararr_iter != this->f90_char_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in charArrays" << endl;
#endif

      const F90_charArray* chararr = (const F90_charArray*)f90chararr_iter->second;

      int ret = IC_Send_char(otherprog, sched, chararr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }
      
    map<string, const F90_shortArray*>::const_iterator f90shortarr_iter; 
    f90shortarr_iter = this->f90_short_arrmap.find(ptname);
    if(f90shortarr_iter != this->f90_short_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in shortArrays" << endl;
#endif
        
      const F90_shortArray* shortarr = (const F90_shortArray*)f90shortarr_iter->second;
        
      int ret = IC_Send_short(otherprog, sched, shortarr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }
      
    map<string, const F90_intArray*>::const_iterator f90intarr_iter; 
    f90intarr_iter = this->f90_int_arrmap.find(ptname);
    if(f90intarr_iter != this->f90_int_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in intArrays" << endl;
#endif
        
      const F90_intArray* intarr = (const F90_intArray*)f90intarr_iter->second;
        
      int ret = IC_Send_int(otherprog, sched, intarr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }
      
    map<string, const F90_floatArray*>::const_iterator f90floatarr_iter; 
    f90floatarr_iter = this->f90_float_arrmap.find(ptname);
    if(f90floatarr_iter != this->f90_float_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in floatArrays" << endl;
#endif
        
      const F90_floatArray* floatarr = (const F90_floatArray*)f90floatarr_iter->second;
        
      int ret = IC_Send_float(otherprog, sched, floatarr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }
      
    map<string, const F90_doubleArray*>::const_iterator f90doublearr_iter; 
    f90doublearr_iter = this->f90_double_arrmap.find(ptname);
    if(f90doublearr_iter != this->f90_double_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in doubleArrays" << endl;
#endif
        
      const F90_doubleArray* doublearr = (const F90_doubleArray*)f90doublearr_iter->second;
        
      int ret = IC_Send_double(otherprog, sched, doublearr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }
#endif
  } 

  if(n_exp == 0){
    cout << "Warning: No export region found with the specified name " << ptname << " in the code" << endl;
  }

  stat = IC_OK;
}

void IC_EndPointSet::importArray(char* ptname, int& stat){
#ifdef IC_DEBUG
  cout << "Import Port" << endl;
#endif
  
  int n_imp = 0;
  map<string, IC_Connection*>::const_iterator citer;

  for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
    string connname = citer->first;    
    IC_Connection* conn = (IC_Connection*)citer->second;
      
    if(!strcmp(conn->get_ctype().c_str(), "1xN")){ 
#ifdef IC_DEBUG
      cout << "\tskip 1xN connection: " << connname << endl;
#endif
      continue;
    }

    if(conn->get_portname().compare(ptname) != 0){ // different portname
#ifdef IC_DEBUG
      cout << "\tskip different port name..." << endl;
#endif
      continue;
    }

    if(conn->get_type() != 1){ // connection is not for export
#ifdef IC_DEBUG
      cout << "\tskip port for export data" << endl;
#endif
      continue;
    }

    n_imp++;

    IC_EndPoint* ep = (IC_EndPoint*)conn->get_ep();
    IC_Program* myprog = ep->get_myprog();
    IC_Program* otherprog = ep->get_otherprog();
    IC_Sched* sched = conn->get_sched();
    int tag = conn->get_tag();

    cout << "\tWaiting to import : " << ptname 
         << " from " << otherprog->name << endl;

#ifdef PPLUSPLUS
    map<string, const intArray*>::const_iterator intarr_iter; 
    intarr_iter = this->ppp_int_arrmap.find(ptname);
    if(intarr_iter != this->ppp_int_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in intArrays" << endl;
#endif
        
      const intArray* intarr = (const intArray*)intarr_iter->second;
        
      int ret = IC_Recv_int(otherprog, sched, 
                            intarr->getLocalArray().getDataPointer(), tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }

    map<string, const floatArray*>::const_iterator floatarr_iter; 
    floatarr_iter = this->ppp_float_arrmap.find(ptname);
    if(floatarr_iter != this->ppp_float_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in floatArrays" << endl;
#endif

      const floatArray* floatarr = (const floatArray*)floatarr_iter->second;

      int ret = IC_Recv_float(otherprog, sched, 
                              floatarr->getLocalArray().getDataPointer(), tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }

    map<string, const doubleArray*>::const_iterator doublearr_iter; 
    doublearr_iter = this->ppp_double_arrmap.find(ptname);
    if(doublearr_iter != this->ppp_double_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in doubleArrays" << endl;
#endif

      const doubleArray* doublearr = (const doubleArray*)doublearr_iter->second;

      int ret = IC_Recv_double(otherprog, sched, 
                               doublearr->getLocalArray().getDataPointer(), tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }
#endif

#ifdef FORTRAN90
    map<string, const F90_charArray*>::const_iterator f90chararr_iter; 
    f90chararr_iter = this->f90_char_arrmap.find(ptname);
    if(f90chararr_iter != this->f90_char_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in charArrays" << endl;
#endif

      const F90_charArray* chararr = (const F90_charArray*)f90chararr_iter->second;

      int ret = IC_Recv_char(otherprog, sched, chararr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }

    map<string, const F90_shortArray*>::const_iterator f90shortarr_iter; 
    f90shortarr_iter = this->f90_short_arrmap.find(ptname);
    if(f90shortarr_iter != this->f90_short_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in shortArrays" << endl;
#endif

      const F90_shortArray* shortarr = (const F90_shortArray*)f90shortarr_iter->second;

      int ret = IC_Recv_short(otherprog, sched, shortarr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }

    map<string, const F90_intArray*>::const_iterator f90intarr_iter; 
    f90intarr_iter = this->f90_int_arrmap.find(ptname);
    if(f90intarr_iter != this->f90_int_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in intArrays" << endl;
#endif

      const F90_intArray* intarr = (const F90_intArray*)f90intarr_iter->second;

      int ret = IC_Recv_int(otherprog, sched, intarr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }

    map<string, const F90_floatArray*>::const_iterator f90floatarr_iter; 
    f90floatarr_iter = this->f90_float_arrmap.find(ptname);
    if(f90floatarr_iter != this->f90_float_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in floatArrays" << endl;
#endif

      const F90_floatArray* floatarr = (const F90_floatArray*)f90floatarr_iter->second;

      int ret = IC_Recv_float(otherprog, sched, floatarr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }

    map<string, const F90_doubleArray*>::const_iterator f90doublearr_iter; 
    f90doublearr_iter = this->f90_double_arrmap.find(ptname);
    if(f90doublearr_iter != this->f90_double_arrmap.end()){
#ifdef IC_DEBUG
      cout << "\t" << ptname << " is found in doubleArrays" << endl;
#endif

      const F90_doubleArray* doublearr = (const F90_doubleArray*)f90doublearr_iter->second;

      int ret = IC_Recv_double(otherprog, sched, doublearr->base, tag);
      if (ret!=0) {
#ifdef IC_DEBUG
        cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
      }
    }

#endif
  }

  if(n_imp == 0){
    cout << "Error: No import region found with the specified name in the code" << endl;
    stat = IC_EndPoint::IC_COMM_FAILURE;
    return;
  }

  stat = IC_OK;
}

void IC_EndPointSet::printErrorMessage(const char* msg, const int ic_err) {
  unsigned nerrs=sizeof(IC_epset_errors)/sizeof(char*);
  if (-ic_err<0 || -ic_err>nerrs) {
    std::cerr << "invalid error code\n";
    return;
  }
  if (msg)
    std::cerr << msg << ": " << IC_epset_errors[-ic_err] << std::endl;
  else
    std::cerr << IC_epset_errors[-ic_err] << std::endl;
}

