/**
 * This class is encapsulation of multiple ICEndPoint instances. 
 * It parses the XJD file and maintain list of ICEndPoint, and encapsulates
 * exportArray and importArray. Thereby, XJD can be used for high-level
 * API.
 * @file IC_EndPointSet.h
 * @author Il-Chul Yoon (iyoon@cs.umd.edu)
 *         University of Maryland, College Park
 */
 
#ifndef IC_ENDPOINTSET_H
#define IC_ENDPOINTSET_H

#ifndef IC_ENDPOINT_H
#include "IC_EndPoint.h"
#endif

#ifndef IC_CONNECTION_H
#include "IC_Connection.h"
#endif

#include "../ezxml/ezxml.h"
#include <map>
#include <string>

/* reserved define for ezxml */
#define XJDVERSION "version"
#define COMPLIST "components"
#define COMP     "component"
#define CONNLIST "connections"
#define CONN     "connection"
/* reserved define for IC_XJDRegion */
#define CONNID   "id"
#define EXPORTER "exporter"
#define IMPORTER "importer"
#define EXPPORT  "exportport"
#define IMPPORT  "importport"
#define DATATYPE "type"
#define TAG      "msgtag"
#define COMMTYPE "commtype"
/* reserved define for IC_XJDComponent */
#define COMPID   "id"
#define COMPNAME "name"
#define N_NODE   "nNode"
#define ARRAYORDER "arrayorder"

using namespace std;


/*!
 This class is encapsulation of multiple ICEndPoint instances. 
 It parses the XJD file and maintain list of ICEndPoint, and encapsulates
 exportArray and importArray. Thereby, XJD can be used for high-level
 API.
*/
class IC_EndPointSet
{
  public:

  /** 
   * Enumeration for InterComm high-level API errors (Specifically for XJD)
   */

  /* the declarations below must be in sync with
     ../FortranInterface/intercomm_interface.h */

    enum
      {
      IC_OK                      = 0,/**<no errors */
      IC_EPSET_INIT_FAILURE      = -13,/**<fail to initialize EndPointSet*/
      IC_ARRAY_REGISTER_FAILURE  = -14,/**<fail to register an array */
      IC_COMMIT_ARRAY_FAILURE    = -15,/**<fail to commit the registered arrays */
    };
    static const char * const IC_epset_errors[];

    /**
     * This will create multiple IC_EndPoint objects related with compname by 
     * parsing the XJD file
     * @param xjdfile the XJD file name
     * @param compname my program name
     * @param status the result of the set creation operation.
     */
    IC_EndPointSet(const char* xjdfile, const char* compname, int& status);

    /**
     * The destructor of the IC_EndPointSet
     */
    virtual ~IC_EndPointSet();

    /**
     * This computes the schedules for all regions defined with this compname
     * @param status returns the result of the commit operation
     */
    void commitArrays(int& status);

    /**
     * Export the arrays after creating the communication endpoints
     * @param arrname the array name
     * @param status returns the result of the export operation
     */
    void exportArray(char* arrname, int& status);

    /**
     * Import the arrays after creating the communication endpoints
     * @param arrname the array name
     * @param status returns the result of the export operation
     */
    void importArray(char* arrname, int& status);

    static void printErrorMessage(const char* msg, const int ic_err);

#ifdef PPLUSPLUS
  private:
    /* maps portname onto P++ array*/
    map<string, const intArray*> ppp_int_arrmap; /* map portname : intArray*/
    map<string, const floatArray*> ppp_float_arrmap; /* map portname : floatArray*/
    map<string, const doubleArray*> ppp_double_arrmap; /* map portname : doubleArray*/

  public:  
    virtual void registerArray(const char*, const intArray&, int& );
    virtual void bcastLocalArray(const char*, const intArray&, int, int&);
    virtual void recvLocalArray(const char*, const intArray&, int, int&);

    virtual void registerArray(const char*, const floatArray&, int& );
    virtual void bcastLocalArray(const char*, const floatArray&, int, int&);
    virtual void recvLocalArray(const char*, const floatArray&, int, int&);

    virtual void registerArray(const char*, const doubleArray&, int& );
    virtual void bcastLocalArray(const char*, const doubleArray&, int, int&);
    virtual void recvLocalArray(const char*, const doubleArray&, int, int&);
#endif

#ifdef FORTRAN90
  private:
    /* maps portname onto F90 arraydesc*/
    map<string, const F90_charArray*>   f90_char_arrmap;
    map<string, const F90_shortArray*>  f90_short_arrmap;
    map<string, const F90_intArray*>    f90_int_arrmap;
    map<string, const F90_floatArray*>  f90_float_arrmap;
    map<string, const F90_doubleArray*> f90_double_arrmap;

  public:  
    virtual void registerArray(const char* , const F90_charArray& , int& );
    virtual void bcastLocalArray(const char*, const F90_charArray&, int, int&);
    virtual void recvLocalArray(const char*, const F90_charArray&, int, int&);

    virtual void registerArray(const char* , const F90_shortArray& , int& );
    virtual void bcastLocalArray(const char*, const F90_shortArray&, int, int&);
    virtual void recvLocalArray(const char*, const F90_shortArray&, int, int&);

    virtual void registerArray(const char* , const F90_intArray& , int& );
    virtual void bcastLocalArray(const char*, const F90_intArray&, int, int&);
    virtual void recvLocalArray(const char*, const F90_intArray&, int, int&);

    virtual void registerArray(const char* , const F90_floatArray& , int& );
    virtual void bcastLocalArray(const char*, const F90_floatArray&, int, int&);
    virtual void recvLocalArray(const char*, const F90_floatArray&, int, int&);

    virtual void registerArray(const char* , const F90_doubleArray& , int& );
    virtual void bcastLocalArray(const char*, const F90_doubleArray&, int, int&);
    virtual void recvLocalArray(const char*, const F90_doubleArray&, int, int&);
   
#endif

 private:
    const char* myprogname;
    char* cid;
    unsigned mynproc; 
    unsigned myao;
    
    char* xjdversion;

    map<string, IC_EndPoint*> epmap; /* map endpoint name : endpoint*/
    map<string, IC_Connection*> connmap; /* map connection name : connection */
    
    /* map<string, IC_EndPoint*> port2ep; /* map portname : endpoint*/
    /* map<string, IC_Desc*> port2desc; /* map portname : descriptor*/
    /* map<string, IC_Region**> port2regionset; /* map portname : regionset*/
    /* map<string, IC_Sched*> port2sched; /* map portname : schedule*/
    /* map<string, int> port2tag; /* map portname : tag*/
    
    void addep(const char* myname, unsigned mynproc, unsigned myao,
               char* othername, unsigned onproc, unsigned oao,
               char* portname, char* commtype, char* datatype, int tag,
               int ex_im, char* connname);
               
    IC_EndPoint* findep(string epname);
    void registerPort(IC_EndPoint* ep, char* ptname, char* ctype,
                      char* dtype, int tag, int& stat);
    void registerConn(IC_EndPoint* ep, char* connname, char* ptname, char* ctype,
                      char* dtype, int tag, int ex_im, int& stat);


    /* Methods for parsing*/
    ezxml_t read_xjd(const char* xjdname);
    char* get_version(ezxml_t xmlRoot);
    int epset_init_1_5(ezxml_t xmlRoot, int& stat);
    int epset_init_2_0(ezxml_t xmlRoot, int& stat);
    int register_program(ezxml_t xmlRoot);
    int inspect_connections(ezxml_t xmlRoot);

    int XJD_Validate(ezxml_t xmlRoot);
    int XJD_Validate_1_5(ezxml_t xmlRoot);
    int XJD_Validate_2_0(ezxml_t xmlRoot);
    int check_multi_import(ezxml_t xmlRoot);

    ezxml_t get_first_comp(ezxml_t xmlRoot);
    ezxml_t get_first_conn(ezxml_t xmlRoot);
    char* get_field_byID(char* id, ezxml_t xmlRoot, char* fieldname);
};

#endif /*_IC_ENDPOINTSET_H_*/
